import copy
import inspect
import simpy

import simpype
import simpype.build


class PropertyDict(dict):
	def __init__(self, sim):
		assert isinstance(sim, simpype.Simulation)
		super().__init__()
		self.sim = sim
		self.env = sim.env
	
	def __setitem__(self, key, value):
		if isinstance(value, Property):
			super().__setitem__(key, value)
		else:
			super().__setitem__(key, Property(self.sim, key, value))


class Property:
	def __init__(self, sim, name, value):
		assert isinstance(sim, simpype.Simulation)
		self.sim = sim
		self.env = sim.env
		self.name = name
		if isinstance(value, dict) and [f for f in value.values() if inspect.isfunction(f)]:
			self._random = simpype.Random(self.sim, value)
			self._value = self._random.value
		else:
			self._random = None
			self._value = value

	@property
	def value(self):
		return self._value

	def copy(self):
		property = Property(self.sim, self.name, self._value)
		property._random = self._random
		return property

	def refresh(self):
		if isinstance(self._random, simpype.Random):
			self._value = self._random.value


class Timestamp:
	def __init__(self, message, timestamp, resource, event):
		assert isinstance(message, Message)
		assert isinstance(resource, simpype.Resource)
		self.message = message
		self.timestamp = float(timestamp)
		self.resource = resource
		self.event = str(event)


class Subscription:
	def __init__(self, sim, message, event, callback, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(message, Message)
		assert callable(callback)
		self.sim = sim
		self.env = sim.env
		self.message = message
		self.event = event
		self.callback = callback
		self.id = id
		self.disable = self.env.event()


class Message:
	def __init__(self, sim, resource, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(resource, simpype.Resource)
		self.sim = sim					
		self.env = sim.env
		self.id = id
		self.generated = self.env.now
		self.generator = resource
		self.is_alive = True
		self.next = {}
		self.visited = []
		self.resource = resource
		self.pipeline = simpype.Pipeline(self.sim, self.id)
		self.property = PropertyDict(self.sim)
		self.location = resource
		self.seq_num = 0
		self.subscription = {}

	@property
	def next(self):
		return self._next
	
	@next.setter
	def next(self, value):
		assert isinstance(value, (simpype.Resource, simpype.Pipeline, dict))
		if isinstance(value, simpype.Resource):
			self._next = {value.id: value}
		elif isinstance(value, simpype.Pipeline):
			self._next = {value.first.id: value.first}
		elif isinstance(value, dict):
			self._next = value

	@property
	def pipeline(self):
		return self._pipeline
	
	@pipeline.setter
	def pipeline(self, value):
		assert isinstance(value, simpype.Pipeline)
		self._pipeline = value
		if hasattr(self, 'resource'):
			self._update_next()

	@property
	def resource(self):
		return self._resource
	
	@resource.setter
	def resource(self, value):
		assert isinstance(value, simpype.Resource)
		self._resource = value
		self.visited.append(value)
		if hasattr(self, 'pipeline'):
			self._update_next()

	def _drop(self, message, cause):
		self.done()
		self.location._message_dropped(self, cause)

	def _update_next(self):
		if self.resource.id in self.pipeline.resource:
			self.next = {p.id: p for p in self.pipeline.resource[self.resource.id]}
		else:
			self.next = {}

	def _wait_event(self, subscription):
		value = yield subscription.event | subscription.disable
		if subscription.event in value and self.is_alive:
			subscription.callback(self, value[subscription.event])
		if subscription.id in self.subscription:
			del self.subscription[subscription.id]
	
	def copy(self):
		message = Message(self.sim, self.generator, self.id)
		message.generated = copy.deepcopy(self.generated)
		message.resource = self.resource
		message.location = self.location
		message.seq_num = copy.deepcopy(self.seq_num)
		message.visited = copy.copy(self.visited)
		message.is_alive = copy.deepcopy(self.is_alive)
		message.next = copy.copy(self.next)
		message.pipeline = copy.copy(self.pipeline)
		for p in self.property.values():
			message.property[p.name] = p.copy()
		for id,s in self.subscription.items():
			c = getattr(message, s.callback.__name__) if inspect.ismethod(s.callback) else s.callback
			s = message.subscribe(event = s.event, callback = c, id = id)
		return message

	def done(self):
		self.is_alive = False
		self.next = {}
		e_list = list(self.subscription.keys())
		for id in e_list:
			self.unsubscribe(id)

	def drop(self, id = 'dropped', event = None):
		if event is None:
			self._drop(self, id)
		else:
			e = self.subscribe(event = event, callback = self._drop, id = id)

	def subscribe(self, event, callback, id):
		assert callable(callback)
		if id in self.subscription:
			self.unsubscribe(id)
		s = Subscription(self.sim, self, event, callback, id)
		self.subscription[id] = s
		self.env.process(self._wait_event(s))
		return s

	def timestamp(self, event):
		ts = Timestamp(self, self.env.now, self.resource, event)
		self.sim.log.write(ts)
		return ts

	def unsubscribe(self, id):
		assert id in self.subscription
		self.subscription[id].disable.succeed()
